r"""Ptpython
============

``ptpython`` use a multi mode editor with emacs, vi insert, vi normal modes,
... So we must know current edit mode to decide how to handle key inputs.
``IME`` provide a skeleton for it.
"""

from dataclasses import dataclass

from prompt_toolkit.filters import Condition
from ptpython.repl import PythonRepl


@dataclass
class IME:
    r"""IME is a class to provide basic support for ``ptpython``."""

    repl: PythonRepl
    preedit: str = ""
    is_enabled: bool = False

    def conditional_disable(self) -> None:
        r"""Conditional disable.

        :rtype: None
        """

    def conditional_enable(self) -> None:
        r"""Conditional enable.

        :rtype: None
        """

    def filter(self, condition: Condition | None = None) -> Condition:
        r"""Filter.

        :param condition:
        :type condition: Condition | None
        :rtype: Condition
        """

        @Condition
        def _(condition: Condition | None = condition) -> bool:
            r""".

            :param condition:
            :type condition: Condition | None
            :rtype: bool
            """
            if condition is None:
                return self.preedit == ""
            return self.preedit == "" and condition()

        return _

    def mode(self, keys: list[str]) -> Condition:
        r"""Mode.

        :param keys:
        :type keys: list[str]
        :rtype: Condition
        """

        @Condition
        def _(keys: list[str] = keys) -> bool:
            r""".

            :param keys:
            :type keys: list[str]
            :rtype: bool
            """
            if len(keys) == 1 == len(keys[0]):
                return self.is_enabled
            elif len(keys) == 1 or len(keys) > 1 and keys[0] == "escape":
                return self.preedit != ""
            else:
                raise NotImplementedError

        return _



"""
Gdb plugin for ptpython. Run
``gdb -x /the/path/of/pyrime/prompt_toolkit/gdb.py`` and type `ptpython`.

Refer `<https://github.com/prompt-toolkit/ptpython/issues/546>`_
"""

import os
import sys
from argparse import Namespace

import gdb  # type: ignore
from ptpython.entry_points.run_ptpython import get_config_and_history_file
from ptpython.repl import PythonRepl, embed, run_config

import __main__


class PtPythonCommand(gdb.Command):
    def __init__(self) -> None:
        super().__init__("ptpython", gdb.COMMAND_USER)
        a = Namespace(config_file=None, history_file=None)
        self.config_file, self.history_file = get_config_and_history_file(a)
        self.startup_paths = []
        if "PYTHONSTARTUP" in os.environ:
            self.startup_paths.append(os.environ["PYTHONSTARTUP"])

    def configure(self, repl: PythonRepl) -> None:
        if os.path.exists(self.config_file):
            run_config(repl, self.config_file)

    def invoke(self, argument: str, from_tty: bool):
        self.dont_repeat()

        if not from_tty:
            raise Exception("PtPython can only be launched from the TTY")

        stdout = sys.stdout
        stderr = sys.stderr
        stdin = sys.stdin

        try:
            sys.stdout = sys.__stdout__
            sys.stderr = sys.__stderr__
            sys.stdin = sys.__stdin__

            embed(  # type: ignore
                history_filename=self.history_file,
                configure=self.configure,
                locals=__main__.__dict__,
                globals=__main__.__dict__,
                startup_paths=self.startup_paths,
                title="GDB REPL (ptpython)",
            )

        except SystemExit as e:
            if e.code != 0:
                print("ptpython exited with code", e.code)

        finally:
            sys.stdout = stdout
            sys.stderr = stderr
            sys.stdin = stdin


PtPythonCommand()



"""cython file"""

import cython as c
from cython.cimports.rime_api import (
    RimeApi,
    RimeCandidate,
    RimeCommit,
    RimeComposition,
    RimeContext,
    RimeMenu,
    RimeSchemaList,
    RimeSchemaListItem,
    RimeTraits,
    rime_get_api,
)

from . import Candidate, Commit, Composition, Context, Menu, SchemaListItem

rime: RimeApi = c.declare(RimeApi)

def init(
    shared_data_dir: str = "/usr/share/rime-data",
    user_data_dir: str = "/root/.config/ibus/rime",
    log_dir: str = "/root/.local/share/ptpython/rime",
    distribution_name: str = "Rime",
    distribution_code_name: str = "pyrime",
    distribution_version: str = "0.0.1",
    app_name: str = "rime.pyrime",
    min_log_level: int = 3,
) -> None:
    r"""Init.

    :param shared_data_dir:
    :type shared_data_dir: str
    :param user_data_dir:
    :type user_data_dir: str
    :param log_dir:
    :type log_dir: str
    :param distribution_name:
    :type distribution_name: str
    :param distribution_code_name:
    :type distribution_code_name: str
    :param distribution_version:
    :type distribution_version: str
    :param app_name:
    :type app_name: str
    :param min_log_level:
    :type min_log_level: int
    :rtype: None
    """
    _min_log_level: c.int = min_log_level
    traits: RimeTraits = RimeTraits(
        shared_data_dir=shared_data_dir.encode(),
        user_data_dir=user_data_dir.encode(),
        distribution_name=distribution_name.encode(),
        distribution_code_name=distribution_code_name.encode(),
        distribution_version=distribution_version.encode(),
        app_name=app_name.encode(),
        min_log_level=_min_log_level,
        log_dir=log_dir.encode(),
    )
    traits.data_size = c.sizeof(RimeTraits) - c.sizeof(traits.data_size)
    global rime
    rime: RimeApi = rime_get_api()[0]
    rime.setup(c.address(traits))
    rime.initialize(c.address(traits))

def create_session() -> int:
    r"""Create session.

    :rtype: int
    """
    return rime.create_session()

def destroy_session(session_id: int) -> None:
    r"""Destroy session.

    :param session_id:
    :type session_id: int
    :rtype: None
    """
    return rime.destroy_session(session_id)

def get_current_schema(session_id: int) -> str:
    r"""Get current schema.

    :param session_id:
    :type session_id: int
    :rtype: str
    """
    schema_id: c.char[1024] = c.declare(c.char[1024])  # type: ignore
    rime.get_current_schema(session_id, schema_id, c.sizeof(schema_id))
    return schema_id.decode()

def get_schema_list() -> list[SchemaListItem]:
    r"""Get schema list.

    :rtype: list[SchemaListItem]
    """
    schema_list: RimeSchemaList = c.declare(RimeSchemaList)
    rime.get_schema_list(c.address(schema_list))
    results: list[SchemaListItem] = []
    i: c.int
    for i in range(schema_list.size):
        schema: RimeSchemaListItem = schema_list.list[i]
        results += [
            SchemaListItem(
                schema.schema_id.decode(),
                schema.name.decode(),
            )
        ]
    return results

def select_schema(session_id: int, schema_id: str) -> bool:
    r"""Select schema.

    :param session_id:
    :type session_id: int
    :param schema_id:
    :type schema_id: str
    :rtype: bool
    """
    return rime.select_schema(session_id, schema_id.encode()) == 1

def process_key(session_id: int, keycode: int, mask: int) -> bool:
    r"""Process key.

    :param session_id:
    :type session_id: int
    :param keycode:
    :type keycode: int
    :param mask:
    :type mask: int
    :rtype: bool
    """
    _keycode: c.int = keycode
    _mask: c.int = mask
    return rime.process_key(session_id, _keycode, _mask) == 1

def get_context(session_id: int) -> Context | None:
    r"""Get context.

    :param session_id:
    :type session_id: int
    :rtype: Context | None
    """
    context: RimeContext = c.declare(RimeContext)
    context.data_size = c.sizeof(RimeContext) - c.sizeof(context.data_size)
    if rime.get_context(session_id, c.address(context)) != 1:
        return None
    composition: RimeComposition = context.composition
    preedit: str | None = (
        None if composition.preedit == c.NULL else composition.preedit.decode()
    )
    menu: RimeMenu = context.menu
    select_keys: str | None = (
        None if menu.select_keys == c.NULL else menu.select_keys.decode()
    )
    candidates: list[Candidate] = []
    i: c.int
    for i in range(menu.num_candidates):
        candidate: RimeCandidate = menu.candidates[i]
        comment: str | None = (
            None if candidate.comment == c.NULL else candidate.comment.decode()
        )
        candidates += [
            Candidate(
                candidate.text.decode(),
                comment,
            )
        ]
    return Context(
        Composition(
            c.cast(int, composition.length),
            c.cast(int, composition.cursor_pos),
            c.cast(int, composition.sel_start),
            c.cast(int, composition.sel_end),
            preedit,
        ),
        Menu(
            c.cast(int, menu.page_size),
            c.cast(int, menu.page_no),
            menu.is_last_page == 1,
            c.cast(int, menu.highlighted_candidate_index),
            c.cast(int, menu.num_candidates),
            select_keys,
            candidates,
        ),
    )

def get_commit(session_id: int) -> Commit | None:
    r"""Get commit.

    :param session_id:
    :type session_id: int
    :rtype: Commit
    """
    commit: RimeCommit = c.declare(RimeCommit)
    commit.data_size = c.sizeof(RimeCommit) - c.sizeof(commit.data_size)
    if rime.get_commit(session_id, c.address(commit)) != 1:
        return None
    return Commit(commit.text.decode())

def commit_composition(session_id: int) -> bool:
    r"""Commit composition.

    :param session_id:
    :type session_id: int
    :rtype: bool
    """
    return rime.commit_composition(session_id) == 1

def clear_composition(session_id: int) -> None:
    r"""Clear composition.

    :param session_id:
    :type session_id: int
    :rtype: None
    """
    rime.clear_composition(session_id)



r"""Plugins
===========

In some python environments such as ``gdb``, 3rd binary python module cannot be
imported. Use ``IME`` to replace ``RIME``.
"""

from .. import IME

try:
    from ..rime import RIME
except ImportError:
    RIME = IME



r"""Test terminfo."""

from pyrime.terminfo import Key, ModifierKey


class Test:
    r"""Test."""

    @staticmethod
    def test_parse_key() -> None:
        r"""Test parse key.

        :rtype: None
        """
        key = Key.new(["c-^"])
        assert key.basic.value == ord("6")
        assert key.modifier == ModifierKey.Control

    @staticmethod
    def test_get_ansi() -> None:
        r"""Test ANSI escape code.

        :rtype: None
        """
        assert (
            ModifierKey.Control | ModifierKey.Shift | ModifierKey.Alt
        ).get_ansi() == 8
        assert (
            ModifierKey.Control | ModifierKey.Shift | ModifierKey.Alt
        ) == ModifierKey.from_ansi(8)

    @staticmethod
    def test_ansi() -> None:
        r"""Test ANSI escape code.

        :rtype: None
        """
        key = Key.new(
            "enter", ModifierKey.Control | ModifierKey.Shift | ModifierKey.Alt
        )
        assert key.get_prompt_toolkit() == ["escape", *"[27;8;13~"]
        assert key == Key.new(["escape", *"[27;8;13~"])



r"""Term info
=============

Convert rime/prompt-toolkit.

Rime uses a key code and a key mask to describe a key. Such as: ``(32, 5)``,
means ``Control(4) + Shift(1) + space(32)``.

However, prompt-toolkit uses ``list[Keys | str]``. ``Keys`` is a ``Enum``. such
as ``[tab, "A"]``, ``KEY_ALIASES.get('tab', 'tab')`` will return
``Keys.ControlI``, means ``Control + I, Shift + A``.

Rime supports more keys, such as ``CapsLock``. Prompt-toolkit supports input
many keys at same time, such as ``Control + I, Shift + A``. We only support
their intersection.
"""

import json
import os
from dataclasses import dataclass
from enum import Enum, Flag, unique
from typing import Self

from prompt_toolkit.keys import KEY_ALIASES, Keys

json_dir = os.path.join(os.path.dirname(__file__), "assets", "json")
with open(os.path.join(json_dir, "keys.json")) as f:
    key_map: dict[str, int] = json.load(f)
with open(os.path.join(json_dir, "modifiers.json")) as f:
    modifiers: list = json.load(f)
key_map = {
    ("_" if k in "".join(map(str, range(10))) else "") + k: v
    for k, v in key_map.items()
}
name_map = {
    "Enter": "Return",
    "Pageup": "Page_Up",
    "Pagedown": "Page_Down",
}
template_map = {
    "Return": "\x1b[27;{};13~",
}


class BasicKeyEnum(Enum):
    r"""BasickeyEnum."""

    def get_rime_name(self) -> str:
        r"""Get rime name.

        :rtype: str
        """
        return {v: k for k, v in key_map.items()}[self.value].lstrip("_")

    def get_pt_name(self) -> str:
        r"""Get pt name.

        :rtype: str
        """
        return self.rime_to_pt(self.get_rime_name())

    def get_template(self) -> str | None:
        r"""Get template.

        :rtype: str | None
        """
        return template_map.get(self.get_rime_name())

    @classmethod
    def from_rime_name(cls, name: str) -> Self:
        r"""From rime name.

        :param name:
        :type name: str
        :rtype: Self
        """
        return cls(
            key_map[
                ("_" if name in "".join(map(str, range(10))) else "") + name
            ]
        )

    @classmethod
    def from_pt_name(cls, name: str) -> Self:
        r"""From pt name.

        :param name:
        :type name: str
        :rtype: Self
        """
        name = cls.pt_to_rime(name)
        return cls.from_rime_name(name)

    @classmethod
    def from_template(cls, template: str) -> Self:
        r"""From template.

        :param template:
        :type template: str
        :rtype: Self
        """
        name = {v: k for k, v in template_map.items()}[template]
        return cls.from_rime_name(name)

    @classmethod
    def new(cls, name_or_template: str) -> Self:
        r"""New.

        :param name_or_template:
        :type name_or_template: str
        :rtype: Self
        """
        if "{}" in name_or_template:
            return cls.from_template(name_or_template)
        elif name_or_template[0].isupper():
            return cls.from_rime_name(name_or_template)
        return cls.from_pt_name(name_or_template)

    @staticmethod
    def rime_to_pt(name: str) -> str:
        r"""Convert rime name to prompt_toolkit name.

        :param name:
        :type name: str
        :rtype: str
        """
        if len(name) == 1:
            return name
        name = {v: k for k, v in name_map.items()}.get(name, name)
        name = name.lower()
        return name

    @staticmethod
    def pt_to_rime(name: str) -> str:
        r"""Convert prompt_toolkit name to rime name.

        :param name:
        :type name: str
        :rtype: str
        """
        if name == " ":
            return "space"
        if len(name) == 1:
            return name
        name = name.capitalize()
        name = name_map.get(name, name)
        if name in key_map:
            return name
        raise NotImplementedError


BasicKey = BasicKeyEnum("BasicKey", key_map)


@unique
class ModifierKey(Flag):
    r"""Not all rime modifier (Lock, ...) are supported in terminfo."""

    NULL = 0
    Shift = 2 ** modifiers.index("Shift")
    Alt = 2 ** modifiers.index("Alt")
    Control = 2 ** modifiers.index("Control")

    def get_ansi(self) -> int:
        r"""Get value for ansi escape code.

        :param self:
        :rtype: int
        """
        d = {v: i for i, v in enumerate(self.__class__.__members__.values())}
        ansi = 1
        for flag in self:
            ansi += 2 ** (d[flag] - 1)
        return ansi

    @classmethod
    def from_ansi(cls, ansi: int) -> Self:
        r"""From ansi.

        :param cls:
        :param ansi:
        :type ansi: int
        :rtype: Self
        """
        d = {i: v for i, v in enumerate(cls.__members__.values())}
        _, _, num = bin(ansi - 1).partition("0b")
        value = cls.NULL
        for n, char in enumerate(list(num)):
            if char == "1":
                value |= d[len(d) - 1 - n]
        return value

    def format(self, template: str) -> str:
        r"""Format.

        :param template:
        :type template: str
        :rtype: str
        """
        return template.format(self.get_ansi())


@dataclass
class Key:
    r"""Key."""

    basic: BasicKey
    modifier: ModifierKey

    def get_rime(self) -> tuple[int, int]:
        r"""Get rime key code and mask.

        :param self:
        :rtype: tuple[int, int]
        """
        return self.basic.value, self.modifier.value

    @classmethod
    def new(
        cls,
        code_or_keys: int | str | list[Keys | str],
        modifier: ModifierKey = ModifierKey.NULL,
    ) -> Self:
        r"""New.

        :param cls:
        :param code_or_keys:
        :type code_or_keys: int | str | list[Keys | str]
        :param modifier:
        :type modifier: ModifierKey
        :rtype: Self
        """
        if isinstance(code_or_keys, list):
            return cls.from_prompt_toolkit(code_or_keys)
        else:
            if isinstance(code_or_keys, int):
                basic = BasicKey(code_or_keys)
            else:
                basic = BasicKey.new(code_or_keys)
            return cls(basic, modifier)

    @classmethod
    def from_rime(cls, code: int, mask: int) -> Self:
        r"""Create a new Key from rime key code and mask.

        :param cls:
        :param code:
        :type code: int
        :param mask:
        :type mask: int
        :rtype: Self
        """
        return cls(BasicKey(code), ModifierKey(mask))

    def get_prompt_toolkit(self) -> list[Keys | str]:
        r"""Get prompt-toolkit key name.

        :param self:
        :rtype: list[Keys | str]
        """
        keys = []
        template = self.basic.get_template()
        if template:
            keys = list(self.modifier.format(template))
            if keys[0] == "\x1b":
                keys[0] = "escape"
        else:
            name = self.basic.get_pt_name()
            for modifier in self.modifier:
                if modifier == ModifierKey.Alt:
                    keys += ["escape"]
                elif modifier == ModifierKey.Shift:
                    name = "s-" + name
                elif modifier == ModifierKey.Control:
                    name = "c-" + name
        return keys

    @classmethod
    def from_prompt_toolkit(cls, keys: list[Keys | str]) -> Self:
        r"""Create a new Key from prompt-toolkit key name.

        :param cls:
        :param keys:
        :type keys: list[Keys | str]
        :rtype: Self
        """
        names: list[str] = []
        modifier = ModifierKey.NULL
        for key in keys:
            if key == "escape":
                if len(keys) == 2:
                    modifier |= ModifierKey.Alt
                else:
                    names += ["\x1b"]
            else:
                names += [key]
        if len(names) == 0:
            raise NotImplementedError
        name = "".join(names)
        if len(names) == 1:
            name = {v: k for k, v in KEY_ALIASES.items()}.get(name, name)
            if name.startswith("c-"):
                modifier |= ModifierKey.Control
                _, _, name = name.partition("c-")
                name = {"^": "6", "-": "_"}.get(name, name)
            if name.startswith("s-"):
                modifier |= ModifierKey.Shift
                _, _, name = name.partition("s-")
            basic = BasicKey.new(name)
        else:
            prefix, _, suffix = name.partition(";")
            modifier |= ModifierKey.from_ansi(int(suffix[0]))
            name = prefix + _ + "{}" + suffix[1:]
            basic = BasicKey.new(name)
        return cls(basic, modifier)



r"""Wrap rime as OOP APIs."""

from dataclasses import dataclass
from typing import Self

from . import Commit, Context, SchemaListItem
from .__main__ import Traits
from ._C import (
    clear_composition,
    commit_composition,
    create_session,
    destroy_session,
    get_commit,
    get_context,
    get_current_schema,
    get_schema_list,
    init,
    process_key,
    select_schema,
)


@dataclass
class Rime:
    r"""Rime is a class to provide OOP APIs for librime."""

    session_id: int = 0

    @staticmethod
    def init(traits: Traits) -> None:
        r"""Init.

        :param traits:
        :type traits: Traits
        :rtype: None
        """
        init(**vars(traits))

    @staticmethod
    def create_session() -> int:
        r"""Create session.

        :rtype: int
        """
        return create_session()

    @classmethod
    def new(cls, traits: Traits) -> Self:
        r"""New.

        :param traits:
        :type traits: Traits
        :rtype: Self
        """
        cls.init(traits)
        return cls(cls.create_session())

    def process_key(self, keycode: int, mask: int) -> bool:
        r"""Process key.

        :param keycode:
        :type keycode: int
        :param mask:
        :type mask: int
        :rtype: bool
        """
        return process_key(self.session_id, keycode, mask)

    def get_context(self) -> Context | None:
        r"""Get context.

        :rtype: Context | None
        """
        return get_context(self.session_id)

    def get_commit(self) -> Commit | None:
        r"""Get commit.

        :rtype: Commit | None
        """
        return get_commit(self.session_id)

    def get_current_schema(self) -> str:
        r"""Get current schema.

        :rtype: str
        """
        return get_current_schema(self.session_id)

    @staticmethod
    def get_schema_list() -> list[SchemaListItem]:
        r"""Get schema list.

        :rtype: list[SchemaListItem]
        """
        return get_schema_list()

    def select_schema(self, schema_id: str) -> bool:
        r"""Select schema.

        :param schema_id:
        :type schema_id: str
        :rtype: bool
        """
        return select_schema(self.session_id, schema_id)

    def commit_composition(self) -> bool:
        r"""Commit composition.

        :rtype: bool
        """
        return commit_composition(self.session_id)

    def clear_composition(self) -> None:
        r"""Clear composition.

        :rtype: None
        """
        clear_composition(self.session_id)

    def __del__(self) -> None:
        r"""Del.

        :rtype: None
        """
        destroy_session(self.session_id)

    def get_commit_text(self) -> str:
        r"""Get commit text.

        :rtype: str
        """
        if self.commit_composition():
            commit = self.get_commit()
            if commit:
                return commit.text
        return ""



r"""Rime for Ptpython
=====================

``RIME`` inherits ``pyrime.ptpython``'s ``IME`` and ``pyrime.rime``'s ``Rime``
to use ``Rime``s OOP APIs to call librime on the basis of ``IME``.
"""

import os
from dataclasses import dataclass

from prompt_toolkit.buffer import Buffer
from prompt_toolkit.key_binding.key_processor import KeyPressEvent
from prompt_toolkit.layout.containers import (
    Float,
    FloatContainer,
    Window,
)
from prompt_toolkit.layout.controls import BufferControl
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.widgets import Frame
from wcwidth import wcswidth

from ..__main__ import Traits
from ..draw_ui import UI, draw_ui
from ..rime import Rime
from ..terminfo import Key, ModifierKey
from . import IME


@dataclass
class RIME(IME, Rime):
    r"""RIME inherit IME and Rime."""

    remember_rime: bool = False
    window: Window = None  # type: ignore
    layout: Layout | None = None
    traits: Traits = None  # type: ignore
    ui: UI = None  # type: ignore
    keys_set: set[tuple[str, ...]] = None  # type: ignore

    def __post_init__(self) -> None:
        r"""Post init.

        :rtype: None
        """
        if self.traits is None:
            self.traits = Traits()
        if self.ui is None:
            self.ui = UI()
        if self.keys_set is None:
            self.keys_set = {
                ("s-tab",),
                ("s-escape",),
                ("escape", "backspace"),
                ("escape", "enter"),
                tuple(
                    Key.new("enter", ModifierKey.Shift).get_prompt_toolkit()
                ),
                tuple(
                    Key.new("enter", ModifierKey.Control).get_prompt_toolkit()
                ),
                tuple(
                    Key.new(
                        "enter", ModifierKey.Shift | ModifierKey.Control
                    ).get_prompt_toolkit()
                ),
                tuple(
                    Key.new(
                        "enter", ModifierKey.Shift | ModifierKey.Alt
                    ).get_prompt_toolkit()
                ),
                tuple(
                    Key.new(
                        "enter", ModifierKey.Control | ModifierKey.Alt
                    ).get_prompt_toolkit()
                ),
                tuple(
                    Key.new(
                        "enter",
                        ModifierKey.Shift
                        | ModifierKey.Control
                        | ModifierKey.Alt,
                    ).get_prompt_toolkit()
                ),
            }
            for order in range(ord(" "), ord("~") + 1):
                self.keys_set |= {(chr(order),)}
            for number in range(1, 24):
                self.keys_set |= {(f"f{number}",)}
            for keyname in {
                "insert",
                "delete",
                "up",
                "down",
                "left",
                "right",
                "home",
                "end",
                "pageup",
                "pagedown",
            }:
                self.keys_set |= {
                    (keyname,),
                    ("c-" + keyname,),
                    ("s-" + keyname,),
                    ("c-s-" + keyname,),
                    ("escape", keyname),
                    ("escape", "c-" + keyname),
                    ("escape", "s-" + keyname),
                    ("escape", "c-s-" + keyname),
                }
            for order in range(ord("@"), ord("[")):
                key = "c-" + chr(order).lower()
                self.keys_set |= {(key,), ("escape", key)}
            self.keys_set |= {("escape", "escape")}
            for order in range(ord("[") + 1, ord("_")):
                key = "c-" + chr(order)
                self.keys_set |= {(key,), ("escape", key)}
            for order in range(ord(" "), ord("~") + 1):
                self.keys_set |= {("escape", chr(order))}

        for keys in self.keys_set:
            keys = list(keys)

            @self.repl.add_key_binding(*keys, filter=self.mode(keys))  # type: ignore
            def _(event: KeyPressEvent, keys: list[str] = keys) -> None:
                r""".

                :param event:
                :type event: KeyPressEvent
                :param keys:
                :type keys: list[str]
                :rtype: None
                """
                self.key_binding(event, keys)

    def key_binding(self, event: KeyPressEvent, keys: list[str]) -> None:
        r"""Key binding.

        :param event:
        :type event: KeyPressEvent
        :param keys:
        :type keys: list[str]
        :rtype: None
        """
        text, lines, col = self.draw(keys)
        self.preedit = lines[0].strip(" " + self.ui.cursor)
        ui_text = "\n".join(lines)
        event.cli.current_buffer.insert_text(text)
        self.window.height = len(lines)
        if len(lines):
            self.window.width = max(wcswidth(line) for line in lines)
        self.window.content.buffer.text = ui_text  # type: ignore
        left, top = self.calculate()
        left += col
        self.repl.app.layout.container.floats[0].left = left  # type: ignore
        self.repl.app.layout.container.floats[0].top = top  # type: ignore

    def draw(self, keys: list[str]) -> tuple[str, list[str], int]:
        r"""Draw.

        :param keys:
        :type keys: list[str]
        :rtype: tuple[str, list[str], int]
        """
        if not self.process_key(*Key.new(keys).get_rime()):
            if len(keys) == 1 == len(keys[0]):
                return keys[0], [self.ui.cursor], 0
            return "", [self.ui.cursor], 0
        context = self.get_context()
        if context is None or context.menu.num_candidates == 0:
            return self.get_commit_text(), [self.ui.cursor], 0
        lines, col = draw_ui(context, self.ui)
        return "", lines, col

    def swap_layout(self):
        r"""Swap layout."""
        (self.layout, self.repl.app.layout) = (  # type: ignore
            self.repl.app.layout,
            self.layout,
        )

    def disable(self) -> None:
        r"""Disable.

        :rtype: None
        """
        self.swap_layout()
        self.is_enabled = False
        self.preedit = ""
        self.clear_composition()

    def calculate(self) -> tuple[int, int]:
        r"""Calculate.

        :rtype: tuple[int, int]
        """
        left = 0
        for _, text in self.repl.all_prompt_styles[  # type: ignore
            self.repl.prompt_style
        ].in_prompt():
            left += wcswidth(text)
        top = 0
        if self.repl.app.layout.current_buffer:
            lines = self.repl.app.layout.current_buffer.text[
                : self.repl.app.layout.current_buffer.cursor_position
            ].split("\n")
            top += len(lines)
            left += wcswidth(lines[-1])
        return left, top

    def enable(self) -> None:
        r"""Enable.

        :rtype: None
        """
        if self.session_id == 0:
            os.makedirs(self.traits.log_dir, exist_ok=True)
            self.init(self.traits)
            self.session_id = self.create_session()
        left, top = self.calculate()
        self.window = Window(
            BufferControl(buffer=Buffer()),
            width=wcswidth(self.ui.cursor),
            height=1,
        )
        self.window.content.buffer.text = self.ui.cursor  # type: ignore
        self.layout = Layout(
            FloatContainer(  # type: ignore
                self.repl.app.layout.container,
                [
                    Float(
                        Frame(
                            self.window,
                        ),
                        left=left,
                        top=top,
                    )
                ],
            )
        )
        self.swap_layout()
        self.is_enabled = True

    def conditional_disable(self) -> None:
        r"""Conditional disable.

        :rtype: None
        """
        if self.is_enabled:
            self.remember_rime = True
            self.disable()

    def conditional_enable(self) -> None:
        r"""Conditional enable.

        :rtype: None
        """
        if self.remember_rime:
            self.enable()

    def toggle(self) -> None:
        r"""Toggle.

        :rtype: None
        """
        if self.is_enabled:
            self.disable()
        else:
            self.enable()



r"""Provide ``__version__`` for
`importlib.metadata.version() <https://docs.python.org/3/library/importlib.metadata.html#distribution-versions>`_.
"""

from dataclasses import dataclass

__version__ = "@PROJECT_VERSION@"


@dataclass
class SchemaListItem:
    r"""Schemalistitem."""

    schema_id: str
    name: str


@dataclass
class Composition:
    r"""Composition."""

    length: int
    cursor_pos: int
    sel_start: int
    sel_end: int
    preedit: str | None


@dataclass
class Candidate:
    r"""Candidate."""

    text: str
    comment: str | None


@dataclass
class Menu:
    r"""Menu."""

    page_size: int
    page_no: int
    is_last_page: bool
    highlighted_candidate_index: int
    num_candidates: int
    select_keys: str | None
    candidates: list[Candidate]


@dataclass
class Context:
    r"""Context."""

    composition: Composition
    menu: Menu


@dataclass
class Commit:
    r"""Commit."""

    text: str
